/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
 * Created on Jan 13, 2005
 *
 */
package com.sunlabs.archsim.gasp;

import com.sunlabs.archsim.base.hierarchy.SimpleEntity;
import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.globals.GlobalDefaults;
import com.sunlabs.archsim.globals.Globals;


/**
 * @author ib27688
 *
 * A terminal forms an interface between one action and one state.
 * A terminal is created when an action and a state are connected.
 * There are three kinds of a terminal:
 * an output terminal, an self resetting input terminal,
 * and a non-resetting input terminal. See SML 2000-0379 for details.
 */
public class Terminal extends SimpleEntity {
	
	protected Globals globals= Globals.getInstance();
	
	// Enum types available only in Java 1.5, this is a workaround
	public static final int OUTPUT= 0;
	public static final int SELF_RESETTING_INPUT= 1;
	public static final int NON_RESETTING_INPUT= 2;
		
	/** Terminal type. There are three types of terminals: 
	 *  output, selfresetting input, non resetting input.
	 *  @see SML 2000-0379 for details.
	 */
	protected int role;
	
	/** pointer to the state */
	protected State attachedState;
	/** pointer to the action */
	protected Action attachedAction;
	
	/** delay for a signal from the action to the state */
	protected Delay actionToStateDelay;
	/** delay for a signal from the state to the action */
	protected Delay stateToActionDelay;

	// protected Command setStateActiveCommand;
	// protected Command setStateInactiveCommand;
	// protected Command notifyActionStateIsActiveCommand;
	// protected Command notifyActionStateIsInactiveCommand;
	
	/** does this terminal need to record its activity? */
	protected boolean journalTerminal;
	
	/** only terminal subclasses can be instantiated */
	protected Terminal(String name) { 
		super(name);
		// setting state and action to null to prepare for a self-check
		attachedState= null;
		attachedAction= null;
		// these attributes need to be set explicitly
		Integer atsDelay= globals.intValue(GlobalDefaults.ACTION_TO_STATE_DELAY);
		actionToStateDelay= (atsDelay != null) 
			? new Delay(atsDelay.intValue())
			: GlobalDefaults.ACTION_TO_STATE_DELAY_DEF;
			
		Integer staDelay= globals.intValue(GlobalDefaults.STATE_TO_ACTION_DELAY)	;
		stateToActionDelay= (staDelay != null)
			? new Delay(staDelay.intValue())
			: GlobalDefaults.STATE_TO_ACTION_DELAY_DEF;
			
		Boolean jTerminal= globals.booleanValue(GlobalDefaults.JOURNAL_TERMINAL);
		journalTerminal= (jTerminal != null)
			? jTerminal.booleanValue()
			: GlobalDefaults.JOURNAL_TERMINAL_DEF;
	} // Terminal
	
	protected Terminal(String name, CompositeEntity g) {
		this(name);
		setParent(g);
	} // Terminal constructor
	
	/** Schedule a notification for an action that the state is active */
	public void notifyStateIsActive() {
		director.schedule(attachedAction.notifyStateIsActiveCommand,
						this, stateToActionDelay);
		if (journalTerminal) {
			director.journal("Terminal " + getAlias() + ", id= " + getID() 
					+ " scheduled notification for action "
					+ attachedAction.getAlias()
					+ ", id= " + attachedAction.getID()
					+ ", that state "
					+ attachedState.getAlias() + 
					", id= " + attachedState.getID() 
					+ " is active");
		}
	} // notifyStateIsActive
	
	/** Schedule a notification for an actionthat a state is inactive */
	public void notifyStateIsInactive() {
		director.schedule(attachedAction.notifyStateIsInactiveCommand,
						this, stateToActionDelay);
		if (journalTerminal) {
			director.journal("Terminal " + getAlias() + ", id= " + getID() 
					+ " scheduled notification for action "
					+ attachedAction.getAlias()
					+ ", id= " + attachedAction.getID()
					+ ", that state "
					+ attachedState.getAlias() + 
					", id= " + attachedState.getID() 
					+ " is INactive");
		}
	} // notifyStateIsInactive
	
	/** Schedule setting the state active
	 * 
	 * @param data the data to be passed to the state
	 * @param d additional dela
	 */
	public void setStateActive(Object data, Delay d) {
		director.schedule(attachedState.setActiveCommand, data, 
				actionToStateDelay.addDelay(d));
		if (journalTerminal) {
			director.journal("Terminal " + getAlias() + ", id= " + getID() 
					+ " scheduled notification for state "
					+ attachedState.getAlias()
					+ ", id= " + attachedState.getID()
					+ ", that action "
					+ attachedAction.getAlias() + 
					", id= " + attachedAction.getID() 
					+ " is making it active with delay"
					+ actionToStateDelay);
		}
	} // setStateActive
	
	/** Schedule setting the state active.
	 * 
	 * @param data data to be passed to the state.
	 */
	public void setStateActive(Object data) {
		director.schedule(attachedState.setActiveCommand, data, 
				actionToStateDelay);
		if (journalTerminal) {
			director.journal("Terminal " + getAlias() + ", id= " + getID() 
					+ " scheduled notification for state "
					+ attachedState.getAlias()
					+ ", id= " + attachedState.getID()
					+ ", that action "
					+ attachedAction.getAlias()
					+ ", id= " + attachedAction.getID() 
					+ " is making it active with delay "
					+ actionToStateDelay + " with data " + data);
		}
	} // setStateActive
	
	
	/** Schedule sttting the state inactive.
	 *
	 */
	public void setStateInactive() {
		director.schedule(attachedState.setInactiveCommand, this, actionToStateDelay);
		if (journalTerminal) {
			director.journal("Terminal " + getAlias() + ", id= " + getID() 
					+ " scheduled notification for state "
					+ attachedState.getAlias()
					+ ", id= " + attachedState.getID()
					+ ", that action "
					+ attachedAction.getAlias() + 
					", id= " + attachedAction.getID() 
					+ " is making it INactive with delay"
					+ actionToStateDelay);
		}
	} // setStateInactive
	
	/* Attach an action to this terminal.
	 */
	public void attachAction(Action a) {
		attachedAction= a;
	} // attachAction

	/* Attach an action to this terminal.
	 */
	public void attachState(State s) {
		attachedState= s;
	} // attachState
	
	/**
	 * set the delay from the action to the state 
	 * @param d delay to be set
	 */
	public void setDelayFromActionToState(Delay d) {
		// clone the delay to avoid inadvertent changes
		actionToStateDelay= (Delay) d.clone();
	} // setDelayFromActionToState

	/**
	 * get the delay from the action to the state 
	 */
	public Delay getDelayFromActionToState() {
		// clone the delay to avoid inadvertent changes
		return (Delay) actionToStateDelay.clone();
	} // getDelayFromActionToState
		
	/**
	 * get the delay from the state to the action 
	 */
	public Delay getDelayFromStateToAction() {
		// clone the delay to avoid inadvertent changes
		return (Delay) stateToActionDelay.clone();
	} // setDelayFromActionToState

	/**
	 * set the delay from the state to the action 
	 * @param d delay to be set
	 */
	public void setDelayFromStateToAction(Delay d) {
		// clone the delay to avoid inadvertent changes
		stateToActionDelay= (Delay) d.clone();
	} // setDelayFromActionToState
	
	
	public State getState() {
		return attachedState;
	} // getState
	
	public Action getAction() {
		return attachedAction;
	} // getAction

	public int getRole() {
		return role;
	} // getRole
	
	public Object getData() {
		return attachedState.getData();
	} // getData
	
	public boolean isAttached(Action a) {
		return (a == attachedAction);
	} // isAttached
	
	public boolean isAttached(State s) {
		return (s == attachedState);
	} // isAttached

	/** A terminal is properly configured and ready for opearation when 
	 * it has attached a state, an action, and it has its delays set.
	 * 
	 * @return true if the terminal is properly configured, false if not
	 */
	public boolean selfCheck() {
		String idStr= "Terminal " + getAlias() 
			+ ", id= " + getID() + ": "; 
		boolean OK= attachedAction != null;
		boolean status=OK;
		if (!OK )
			director.error(idStr + " Not connected to an action");

		OK= attachedState != null;
		status= status && OK;
		if (!OK )
			director.error(idStr + " Not connected to a state");
		
		OK= stateToActionDelay != null;
		status= status && OK;
		if (!OK )
			director.error(idStr + " stateToActionDelay not set");
		
		OK= actionToStateDelay != null;
		status= status && OK;
		if (!OK )
			director.error(idStr + " actionToStateDelay not set");
		
		if (attachedAction != null) {
			OK=  attachedAction.isAttached(this); 
			status= status && OK;
			if (!OK )
				director.error(idStr + " Attached action does not know it is attached to this terminal");
		}
		
		if (attachedState != null) {
			OK= attachedState.isAttached(this);
			status= status && OK;
			if (!OK )
				director.error(idStr + " Attached state does not know it is attached to this terminal");
		}

		return status;
	} // selfCheck
	
	public void print() {
		director.info("Terminal " + getAlias() + ", id= " + getID() 
				+ " is attached to: ");
		if (attachedState!= null) 
			director.info("\t" + "state " + attachedState.getAlias()
						+ ", id= " + attachedState.getID());
		if (attachedAction != null)
			
			director.info("\t" + "action " + attachedAction.getAlias()
						+ ", id= " + attachedAction.getID());
	} // printConnections
		
	
	/** Entity initialization. Do nothing. */
	public void init() {
		
	}
	
} // class Terminal
